
// Ranges and strides, and subscripts and indexes.

#include "RangeRep.h"
#include "CollapsedRangeRep.h"
#include "DimRangeRep.h"
#include "BlockRangeRep.h"
#include "CyclicRangeRep.h"
#include "StepRangeRep.h"

#define MAX_LEVELS 2 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Range': handle class for range representations.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

enum Layout {LAY_STANDARD, LAY_PACKED, LAY_ADAPT, LAY_EXTENDED} ;

class Stride ;
class Subscript ;
class Block ;

class Range {
public :
  // Constructors and destructors

  Range() : rep(0) {}

  Range(const Range& x) : rep(x.rep) {
    if(rep) rep->incr() ;
  }

  Range(const int extent) {
    rep = new CollapsedRangeRep(new DimRangeRep(), extent) ;
    rep->incr() ;
  }

  Range(Dimension dim) {
    rep = new DimRangeRep(dim) ;
    rep->incr() ;
  }

  ~Range() {if(rep) rep->decr() ; }

  // Subranges

  Range subrng(const int extent,
               const int base = 0, const int stride = 1) const {
    if(extent != 0) {
      const int parExtent = size() ;
      const int top = base + stride * (extent - 1) ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: subrng: subrange out of range") ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  // Inquiries

  int size() const {return rep->size() ; }

  Dimension dim() const {return rep->dim() ; }

  int lev() const {return rep->lev() ; }
  Range ker() const { return Range(rep->ker()) ; }

  Range tem() const { return Range(rep->tem()) ; }
  int bas() const {return rep->bas() ; }
  int str() const {return rep->str() ; }

  Format format() const {return rep->format() ; }

  // Subscripting and iteration

  int ker_glb(const int glb, int* sub) const {
    return rep->ker_glb(glb, sub) ;
  }

  int block(const int ker_glb, int* sub, int* count) const {
    return rep->block(ker_glb, sub, count) ;
  }
  int block_step(int* sub_stp) const {
    return rep->block_step(sub_stp) ;
  }

  int crd(const int glb) const {return rep->crd(glb) ; }
  int local(int* glb) const {return rep->local(glb) ; }

  Range subker(int* ker_bas, int* ker_str) const {
    return Range(rep->subker(ker_bas, ker_str)) ;
  }

  // `Stride' stuff

  int width() const {return rep->width() ; }

  int volume() const {
    int size = 1 ;
    for(RangeRep* rng = rep ; rng->lev() ; rng = rng->ker())
      size *= rng->width() ;
    return size ;
  }

  Stride stride(int &size) const ;

  // Miscellaneous operators

  Subscript operator()(const int k) ;

  Range& operator=(const Range& x) {
    if(x.rep) x.rep->incr() ;
    if(rep)   rep->decr() ;
    rep = x.rep ;
    return *this ;
  }

  operator RangeRep*() const {return rep ; }

  // Interim support for PCRC Fortran interface

  int glb(const int sub, const int ker_glb) const {
    return rep->glb(sub, ker_glb) ;
  }

  int& g_lb() {return rep->g_lb ; }
  int& g_stride() {return rep->g_stride ; }

  // debugging

  void show() {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Range :: show: Null range\n") ;
  }
protected :
  Range(RangeRep* _rep) : rep(_rep) {
    if(rep) rep->incr() ;
  }

  RangeRep* rep ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRange': block-distributed range.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRange : public Range {
public :
  BlockRange(Range ker, const int blockSize,
	     const Layout layout = LAY_STANDARD) {
    switch(layout) {
      case LAY_STANDARD :
	rep = new BlockRangeRep(ker, blockSize) ;
	break ;
      case LAY_PACKED :
	rep = new PackedBlockRangeRep(ker, blockSize) ;
	break ;
      default :
	AdlibError("BlockRange :: BlockRange: unsupported layout option\n") ;
    }
    rep->incr() ;
  }

  BlockRange(const int extent, Range ker, const Layout layout = LAY_STANDARD) {
    const int blockSize = (extent + ker.size() - 1) / ker.size() ;
    RangeRep* par ;
    switch(layout) {
      case LAY_STANDARD :
	par = new BlockRangeRep(ker, blockSize) ;
	break ;
      case LAY_PACKED :
	par = new PackedBlockRangeRep(ker, blockSize) ;
	break ;
      default :
	AdlibError("BlockRange :: BlockRange: unsupported layout option\n") ;
    }
    rep = extent != par->size() ? par->subrng(extent, 0, 1) : par ;
    rep->incr() ;
  }

  BlockRange(Range ker, const int blockSize,
	     const Layout layout, const int wlo, const int whi) {
    switch(layout) {
      case LAY_EXTENDED :
	rep = new ExtendedBlockRangeRep(ker, blockSize, wlo, whi) ;
	break ;
      default :
	AdlibError("BlockRange :: BlockRange: unsupported layout option\n") ;
    }
    rep->incr() ;
  }

  BlockRange(const int extent, Range ker,
	     const Layout layout, const int wlo, const int whi) {
    const int blockSize = (extent + ker.size() - 1) / ker.size() ;
    RangeRep* par ;
    switch(layout) {
      case LAY_EXTENDED :
	par = new ExtendedBlockRangeRep(ker, blockSize, wlo, whi) ;
	break ;
      default :
	AdlibError("BlockRange :: BlockRange: unsupported layout option\n") ;
    }
    rep = extent != par->size() ? par->subrng(extent, 0, 1) : par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRange': cyclic-distributed range.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRange : public Range {
public :
  CyclicRange(Range ker, const int numCycles,
	      const Layout layout = LAY_STANDARD) {
    switch(layout) {
      case LAY_STANDARD :
	rep = new CyclicRangeRep(ker, numCycles) ;
	break ;
      case LAY_ADAPT :
	rep = new ADAPTCyclicRangeRep(ker, numCycles) ;
	break ;
      default :
	AdlibError("CyclicRange :: CyclicRange: "
		   "unsupported layout option\n") ;
    }
    rep->incr() ;
  }

  CyclicRange(const int extent, Range ker, const Layout layout = LAY_STANDARD) {
    const int numCycles = (extent + ker.size() - 1) / ker.size() ;
    RangeRep* par ;
    switch(layout) {
      case LAY_STANDARD :
	par = new CyclicRangeRep(ker, numCycles) ;
	break ;
      case LAY_ADAPT :
	par = new ADAPTCyclicRangeRep(ker, numCycles) ;
	break ;
      default :
	AdlibError("CyclicRange :: CyclicRange: "
		   "unsupported layout option\n") ;
    }
    rep = extent != par->size() ? par->subrng(extent, 0, 1) : par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRange': stepped block-distributed range.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRange : public Range {
public :
  StepRange(const int extent, Range ker, const Layout layout = LAY_STANDARD) {
    const int blockSize = (extent + ker.size() - 1) / ker.size() ;
    switch(layout) {
      case LAY_STANDARD :
	rep = new StepRangeRep(ker, extent / ker.size(),
				    extent % ker.size()) ;
	break ;
      case LAY_PACKED :
	rep = new PackedStepRangeRep(ker, extent / ker.size(),
				     extent % ker.size()) ;
	break ;
      default :
	AdlibError("StepRange :: StepRange: unsupported layout\n") ;
    }
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Subscript': The local subscript.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Subscript : public Coord {
  Subscript() {}

  Subscript(const int glb) {

    // Conversion constructor---create subscript in collapsed range.

    subs [0] = glb ;
  }

  Subscript(Range x, int glb) {
    int* sub = subs ;
    RangeRep* rng ;
    for(rng = x ; rng->lev() ; rng = rng->ker())
      glb = rng->ker_glb(glb, sub++) ;

    dim = rng->dim() ;
    crd = rng->crd(glb) ;
  }

  int at() {
    dim.where() ;
    return dim.crd() == crd ;
  }

  void ta() {
    dim.erewh() ;
  }

  int subs [MAX_LEVELS] ;
} ;

inline Subscript Range :: operator()(const int glb) {
  return Subscript(rep, glb) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Block'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Block {
  int count ;

  int glb_bas ;
  int glb_stp ;

  int sub_bas ;
  int sub_stp ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Stride'                                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Stride {
  Stride() : rep(0) {}

  Stride(const Stride& x) : rep(x.rep) {
    if(rep) rep->incr() ;
  }

  ~Stride() {if(rep) rep->decr() ; }

  Stride ker() const {return Stride(rep->ker()) ; }

  int len() const {return rep->len() ; }

  int disp(const int sub) const {return rep ? rep->disp(sub) : 0 ; }

  int disp_step(const int sub_stp) const {
    return rep ? rep->disp_step(sub_stp) : 0 ;
  }

  int offset(const Subscript& i) const {
    int off = 0 ;
    const int* sub = i.subs ;
    for(StrideRep* str = rep ; str ; str = str->ker())
      off += str->disp(*(sub++)) ;
    return off ;
  }

  int step(Block& b) const {
    return disp_step(b.sub_stp) ;
  }

  Stride& operator=(const Stride& x) {
    if(x.rep) x.rep->incr() ;
    if(rep)   rep->decr() ;
    rep = x.rep ;
    return *this ;
  }

  operator StrideRep*() const {return rep ; }

private :
  Stride(StrideRep* _rep) : rep(_rep) {
    if(rep) rep->incr() ;
  }

  StrideRep* rep ;

friend Range ;
} ;

inline Stride Range :: stride(int &size) const {
  if(lev()) {
    const int len = size ;
    size *= width() ;

    return Stride(rep->vector(ker().stride(size), len)) ;
  }
  else
    return Stride() ;
}


// Iterator (index) classes

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlkIt': loop state for scanning individual block.            //
//                                                                     //
//                Note that `glb' is computed relative to a            //
//                parent range (``y'') of `x'.                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlkIt : public Block {
public :
  void init(const Range _x, const int _bas, const int _str) {

    // `_bas', `_str' are alignment parameters of `x' relative to `y'.

    x = _x ;

    bas = _bas ;
    str = _str ;

    glb_stp = str * x.block_step(&sub_stp) ;
  }

  void begin(int* sub, const int ker_glb) {
    glb_bas = bas + str * x.block(ker_glb, &sub_bas, &count) ;

    *sub = sub_bas ;
    glb  = glb_bas ;

    l = 0 ;
  }

  int test() {return l < count ; }

  void next(int* sub) {
    *sub += sub_stp ;
    glb  += glb_stp ;

    l++ ;
  }

  int glb ;

protected :
  Range x ;

  int l ;  // induction variable

  int bas, str ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex': enumerate local blocks of range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex : public Subscript, public Block {
public :
  LocBlocksIndex(Range _x) : x(_x), level(_x.lev()) {

    if(level == 0) {
      glb_stp = 1 ;
      d = x ;
    }
    else {
      glb_stp = x.block_step(&sub_stp) ;
      d = x.subker(&ker_glb_bas, &ker_glb_stp) ;

      for(int r = 1 ; r < level ; r++) {
	blk [r].init(d, ker_glb_bas, ker_glb_stp) ;
	d = d.subker(&ker_glb_bas, &ker_glb_stp) ;
      }
    }

    dim = d.dim() ;
  }

  void beginLocBlk() {
    int subker_glb ;
    if(d.local(&subker_glb)) {
      crd = dim.crd() ;

      switch(level) {
        case 0 :
          glb_bas = subker_glb ;
          count   = 1 ;
          looping = 1 ;

          break ;

        case 1 :
          glb_bas = x.block(ker_glb_bas + ker_glb_stp * subker_glb,
                            &sub_bas, &count) ;
          looping = count ;

          *subs   = sub_bas ;
          break ;

        default : {
          looping = 1 ;
          int r = level - 1 ;
          blk [r].begin(subs + r, ker_glb_bas + ker_glb_stp * subker_glb) ;
          scan(r) ;

          *subs   = sub_bas ;
          break ;
        }
      }
    }
    else
      looping = 0 ;
  }

  int test() {return looping ; }

  void nextLocBlk() {
    if(level < 2)
      looping = 0 ;
    else {
      blk [1].next(subs + 1) ;
      scan(1) ;

      *subs   = sub_bas ;
    }
  }

private :

  void scan(int r) {

    // Starting from current position in r'th kernel range,
    // find first non-empty block.

    // 1 <= r < level

    // (Only used for level >= 2.)

    if(blk [r].test()) {
      const int ker_glb = blk [r].glb ;
      if(r == 1) {
        glb_bas = x.block(ker_glb, &sub_bas, &count) ;

        if(count == 0) {
          blk [1].next(subs + 1) ;
          scan(1) ;
        }
      }
      else {
        r-- ;
        blk [r].begin(subs + r, ker_glb) ;
        scan(r) ;
      }
    }
    else {
      r++ ;
      if(r == level)
        looping = 0 ;
      else {
        blk [r].next(subs + r) ;
        scan(r) ;
      }
    }
  }

protected :
  Range x, d ;
  int level ;

  int ker_glb_bas, ker_glb_stp ;  // These relate to innermost, level 0 kernel

  BlkIt blk [MAX_LEVELS] ;  // blk [0] is unused.
  int looping ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index'.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index : public LocBlocksIndex {
public :
  Index(Range _x) : LocBlocksIndex(_x) {}

  void begin() {
    beginLocBlk() ;
    glb = glb_bas ;

    l   = 0 ;
  }

  void next() {
    if(level == 0)
      nextLocBlk() ;
    else {
      *subs += sub_stp ;
      glb   += glb_stp ;

      l++ ;
      if(l == count) {
        nextLocBlk() ;
        glb = glb_bas ;

        l   = 0 ;
      }
    }
  }

  operator int() const {return glb ; }

private :
  int glb ;
  int l ;

public :

  // Cosmetic support for ghost regions:

  Subscript operator<<(const int shift) {
    Subscript shifted(*this) ;
    *(shifted.subs) -= sub_stp * shift ;
    return shifted ;
  }

  Subscript operator>>(const int shift) {
    Subscript shifted(*this) ;
    *(shifted.subs) += sub_stp * shift ;
    return shifted ;
  }
} ;


// GNU compiler needs these

inline int operator+(const Index& i, const int n) {
  return (int) i + n ;
}

inline int operator+(const int n, const Index& i) {
  return n + (int) i ;
}

inline int operator+(const Index& i, const Index& j) {
  return (int) i + (int) j ;
}

inline int operator-(const Index& i, const int n) {
  return (int) i - n ;
}

inline int operator-(const int n, const Index& i) {
  return n - (int) i ;
}

inline int operator-(const Index& i, const Index& j) {
  return (int) i - (int) j ;
}

inline int operator*(const Index& i, const int n) {
  return (int) i * n ;
}

inline int operator*(const int n, const Index& i) {
  return n * (int) i ;
}

inline int operator*(const Index& i, const Index& j) {
  return (int) i * (int) j ;
}

inline int operator/(const Index& i, const int n) {
  return (int) i / n ;
}

inline int operator/(const int n, const Index& i) {
  return n / (int) i ;
}

inline int operator/(const Index& i, const Index& j) {
  return (int) i / (int) j ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex': enumerate all blocks of range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex : public LocBlocksIndex {
public :
  AllBlocksIndex(Range _x) : LocBlocksIndex(_x) {}

  void beginAllBlk() {
    switch(level) {
      case 0 :
	crd     = d.bas() ;
	glb_bas = 0 ;

	looping = d.size() ;

        count   = 1 ;
	break ;

      case 1 :
	crd     = d.bas() ;
	ker_glb = ker_glb_bas ;

	l       = 0 ;

	looping = d.size() ;
	if(looping) {
	  glb_bas = x.block(ker_glb, &sub_bas, &count) ;
	  while(count == 0) {
	    crd     += d.str() ;
	    ker_glb += ker_glb_stp ;

	    l++ ;
	    if(l == d.size()) {
	      looping = 0 ;
	      break ;
	    }
	    else
	      glb_bas = x.block(ker_glb, &sub_bas, &count) ;
	  }
	  *subs   = sub_bas ;
	}
	break ;

      default : {
	crd     = d.bas() ;
	ker_glb = ker_glb_bas ;

	l       = 0 ;

        looping = 1 ;
	int r = level - 1 ;
	blk [r].begin(subs + r, ker_glb) ;
	scan(r) ;

	*subs   = sub_bas ;
	break ;
      }
    }
  }

  void nextAllBlk() {
    switch(level) {
      case 0 :
	crd     += d.str() ;
	glb_bas++ ;

	looping = glb_bas < d.size() ;
	break ;

      case 1 :
        do {
	  crd     += d.str() ;
	  ker_glb += ker_glb_stp ;

	  l++ ;

	  if(l == d.size()) {
	    looping = 0 ;
	    break ;
	  }
	  else
	    glb_bas = x.block(ker_glb, &sub_bas, &count) ;
        } while (count == 0) ;

	*subs   = sub_bas ;
	break ;

      default :
	blk [1].next(subs + 1) ;
	scan(1) ;

	*subs   = sub_bas ;
	break ;
    }
  }

private :

  void scan(int r) {

    // Starting from current position in r'th kernel range,
    // find first non-empty block.

    // 1 <= r < level

    // (Only used for level >= 2.)

    if(blk [r].test()) {
      const int ker_glb_r = blk [r].glb ;
      if(r == 1) {
        glb_bas = x.block(ker_glb_r, &sub_bas, &count) ;

        if(count == 0) {
          blk [1].next(subs + 1) ;
          scan(1) ;
        }
      }
      else {
        r-- ;
        blk [r].begin(subs + r, ker_glb_r) ;
        scan(r) ;
      }
    }
    else {
      if(r == level - 1) {
        crd     += d.str() ;
        ker_glb += ker_glb_stp ;

        l++ ;
	if(l == d.size())
	  looping = 0 ;
        else {
          blk [r].begin(subs + r, ker_glb) ;
          scan(r) ;
        }
      }
      else {
        r++ ;
        blk [r].next(subs + r) ;
        scan(r) ;
      }
    }
  }

private :

  int l ;
  int ker_glb ;  // This relates to innermost, level 0 kernel
} ;



// ---------------------------------------------------------------
//
// Things to do:
//
//   Change `subrng' to take lower-bound/upper-bound type parametrization???
//       Don't think so.
//
//   Abandon inlining most members of iterators?  The functions are too
//       complex, and confuse compilers (eg, DEC).

